
// =================================================================================================
// This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. It
// is auto-generated by the 'scripts/database/database.py' Python script.
//
// This file populates the database with best-found tuning parameters for the 'Gemm_Routine' kernels.
//
// =================================================================================================

#include "database/database_structure.hpp"

namespace clblast {
namespace database {

extern const DatabaseEntry GemmRoutineHalf;
extern const DatabaseEntry GemmRoutineSingle;
extern const DatabaseEntry GemmRoutineComplexSingle;
extern const DatabaseEntry GemmRoutineDouble;
extern const DatabaseEntry GemmRoutineComplexDouble;

} // namespace database
} // namespace clblast
